/*
  Copyright 2018 Tautvydas Andrikys

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

      http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.esminis.server.mariadb.server;

import android.content.Context;
import android.os.Build;
import android.os.StatFs;

import com.esminis.server.library.model.NetworkInterface;
import com.esminis.server.library.server.ServerControlStartConfig;
import com.esminis.server.library.server.ServerLauncher;
import com.esminis.server.library.service.Utils;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import io.reactivex.Completable;
import io.reactivex.schedulers.Schedulers;

class MariaDbServerLauncher extends ServerLauncher<ServerControlStartConfig> {

	private final Object lock = new Object();

	MariaDbServerLauncher(com.esminis.server.library.model.manager.Process managerProcess) {
		super(managerProcess);
	}

	void stop(Process process, File binary) {
		synchronized (lock) {
			managerProcess.stop(process, binary, 15, null);
		}
	}

	@Override
	protected File getDirectoryWork(ServerControlStartConfig config) {
		return config.root;
	}

	@Override
	protected List<String> createCommand(ServerControlStartConfig config) throws Exception {
		return null;
	}

	Process start(ServerControlStartConfig config, Context context) throws IOException {
		initializeDataDirectory(context, config.binary, config.root);
		synchronized (lock) {
			return start(
				config.binary, createCommand(context, config), createEnvironment(config),
				getDirectoryWork(config)
			);
		}
	}

	private long getFreeSpace(File file) {
		StatFs stat = new StatFs(file.getPath());
		if (Build.VERSION.SDK_INT > Build.VERSION_CODES.JELLY_BEAN_MR2) {
			return stat.getAvailableBlocksLong() * stat.getBlockSizeLong();
		}
		//noinspection deprecation
		return (long)stat.getAvailableBlocks() * (long)stat.getBlockSize();
	}

	private String getVersion(File binary) throws IOException {
		final Process process = Runtime.getRuntime()
			.exec(new String[] {binary.getAbsolutePath(), "--version"});
		try {
			process.waitFor();
		} catch (InterruptedException e) {
			return null;
		}
		final String version = IOUtils.toString(process.getInputStream(), Charset.defaultCharset());
		return version == null || version.isEmpty() ? null : version;
	}

	void initializeDataDirectory(Context context, File binary, File root) throws IOException {
		final File[] files = root.listFiles();
		if (files != null && files.length > 0) {
			return;
		}
		final String version = getVersion(binary);
		synchronized (lock) {
			final List<String> environment = createEnvironment();
			final List<String> command = createCommandInternal(context, binary, root, null);
			final File defaultCnf = getDefaultCnf(root);
			Collections.addAll(
				command, "--bootstrap", "--log-warnings=0", "--max_allowed_packet=8M",
				"--net_buffer_length=16K"
			);
			final File dataMysqlDirectory = new File(root, "mysql");
			if (dataMysqlDirectory.isDirectory()) {
				return;
			}
			Utils.createDirectory(dataMysqlDirectory);
			final Process process = Runtime.getRuntime()
				.exec(command.toArray(new String [0]), environment.toArray(new String[0]), root);
			final Object[] finishedWithError = {null};
			final String[] log = {""};
			try {
				final OutputStream stream = process.getOutputStream();
				Completable.create(
					emitter -> {
						final InputStream inputStream = process.getErrorStream();
						for (;;) {
							synchronized (finishedWithError) {
								if (finishedWithError[0] != null) {
									break;
								}
							}
							try {
								int available = inputStream.available();
								if (available > 0) {
									for (int i = 0; i < available; i++) {
										log[0] += (char)inputStream.read();
									}
									if (
										getFreeSpace(dataMysqlDirectory) < 1024L * 1024L ||
										log[0].contains("No space left on device")
									) {
										synchronized (finishedWithError) {
											finishedWithError[0] = new IOException("No space left on device");
										}
										process.destroy();
										break;
									}
								}
							} catch (Throwable ignored) {}
							Thread.yield();
						}
						emitter.onComplete();
					}
				).subscribeOn(Schedulers.newThread()).subscribe();
				writeToStream(stream, "use mysql;\n");
				if (version != null && version.contains("10.0.0")) {
					writeToStream(stream, context, "sql/mysql_system_tables_10_0.sql");
					writeToStream(stream, context, "sql/mysql_system_tables_data_10_0.sql");
					writeToStream(stream, context, "sql/add_root_from_any_host_10_0.sql");
				} else {
					if (version != null && version.contains("10.3.")) {
						writeToStream(stream, context, "sql/mysql_system_tables_10_3.sql");
					} else {
						writeToStream(stream, context, "sql/mysql_system_tables.sql");
					}
					writeToStream(stream, context, "sql/mysql_performance_tables.sql");
					writeToStream(stream, context, "sql/mysql_system_tables_data.sql");
					writeToStream(stream, context, "sql/add_root_from_any_host.sql");
				}
				writeToStream(stream, context, "sql/fill_help_tables.sql");
				writeToStream(stream, "exit;\n");
				process.waitFor();
				if (!defaultCnf.isFile()) {
					try {
						//noinspection ResultOfMethodCallIgnored
						defaultCnf.createNewFile();
					} catch (IOException ignored) {}
				}
			} catch (Throwable e) {
				FileUtils.deleteDirectory(root);
				//noinspection ResultOfMethodCallIgnored
				root.mkdirs();
				synchronized (finishedWithError) {
					if (finishedWithError[0] != null && finishedWithError[0] instanceof IOException) {
						throw (IOException)finishedWithError[0];
					}
					throw new IOException(
						e.toString() + "\n\nLog:\n" +
							IOUtils.toString(process.getErrorStream(), Charset.defaultCharset()) + "," +
							log[0] + IOUtils.toString(process.getInputStream(), Charset.defaultCharset())
					);
				}
			} finally {
				synchronized (finishedWithError) {
					if (finishedWithError[0] == null) {
						finishedWithError[0] = true;
					}
				}
			}
		}
	}

	private void writeToStream(OutputStream output, Context context, String path) throws IOException {
		IOUtils.copy(context.getAssets().open(path), output);
		output.flush();
	}

	private void writeToStream(OutputStream outputStream, String content) throws IOException {
		outputStream.write(content.getBytes());
		outputStream.flush();
	}

	private List<String> createCommandInternal(
		Context context, File binary, File root, File myCnf
	) throws IOException {
		final List<String> command = new ArrayList<>();
		final File fileSocket = new File(binary.getParentFile(), "mysql.sock");
		command.add(binary.getAbsolutePath());
		final boolean hasLowerCaseTableNames;
		if (myCnf != null && myCnf.isFile()) {
			command.add("--defaults-file=" + myCnf.getAbsolutePath());
			hasLowerCaseTableNames = FileUtils.readFileToString(myCnf, Charset.defaultCharset())
				.contains("lower_case_table_names");
		} else {
			hasLowerCaseTableNames = false;
		}
		if (!hasLowerCaseTableNames) {
			command.add("--lower-case-table-names=1");
		}
		command.add("--lc-messages-dir=" + new File(binary.getParentFile(), "share"));
		command.add(
			"--tmpdir=" +
			Utils.createDirectory(new File(context.getExternalFilesDir(null), "temp")).getAbsolutePath()
		);
		command.add("--socket=" + fileSocket);
		command.add("--basedir=" + binary.getParentFile().getAbsolutePath());
		command.add("--datadir=" + root.getAbsolutePath());
		return command;
	}

	private List<String> createCommand(
		Context context, ServerControlStartConfig config
	) throws IOException {
		final List<String> command = createCommandInternal(
			context, config.binary, config.root, getDefaultCnf(config.root)
		);
		if (config.address.networkInterface != NetworkInterface.ALL) {
			command.add("--bind-address=" + config.address.networkInterface.getIpAddress());
		}
		command.add("--port=" + config.address.port);
		command.add("--pid-file=" + new File(config.binary.getParent(), "pid").getAbsolutePath());
		return command;
	}

	private File getDefaultCnf(File directory) {
		return new File(directory, "my.cnf");
	}

}
